"""
$description Live streaming video game broadcasts from the Steam gaming community.
$url steamcommunity.com
$url steam.tv
$type live
$account Some streams require a login
"""

import base64
import logging
import re
import time

from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA

from streamlink.exceptions import FatalPluginError
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream

log = logging.getLogger(__name__)


class SteamLoginFailed(Exception):
    pass


@pluginmatcher(re.compile(
    r"https?://steamcommunity\.com/broadcast/watch/(\d+)"
))
@pluginmatcher(re.compile(
    r"https?://steam\.tv/(\w+)"
))
@pluginargument(
    "email",
    requires=["password"],
    metavar="EMAIL",
    help="A Steam account email address to access friends/private streams",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="A Steam account password to use with --steam-email.",
)
class SteamBroadcastPlugin(Plugin):
    _watch_broadcast_url = "https://steamcommunity.com/broadcast/watch/{steamid}"
    _get_broadcast_url = "https://steamcommunity.com/broadcast/getbroadcastmpd/"
    _get_rsa_key_url = "https://steamcommunity.com/login/getrsakey/"
    _dologin_url = "https://steamcommunity.com/login/dologin/"
    _captcha_url = "https://steamcommunity.com/public/captcha.php?gid={}"

    @property
    def donotcache(self):
        return str(int(time.time() * 1000))

    def encrypt_password(self, email, password):
        """
        Get the RSA key for the user and encrypt the user's password
        :param email: steam account
        :param password: password for account
        :return: encrypted password
        """
        rsadata = self.session.http.get(
            self._get_rsa_key_url,
            params=dict(
                username=email,
                donotcache=self.donotcache,
            ),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "publickey_exp": validate.all(str, validate.transform(lambda x: int(x, 16))),
                    "publickey_mod": validate.all(str, validate.transform(lambda x: int(x, 16))),
                    "success": True,
                    "timestamp": str,
                    "token_gid": str
                }
            )
        )

        rsa = RSA.construct((rsadata["publickey_mod"], rsadata["publickey_exp"]))
        cipher = PKCS1_v1_5.new(rsa)
        return base64.b64encode(cipher.encrypt(password.encode("utf8"))), rsadata["timestamp"]

    def dologin(self, email, password, emailauth="", emailsteamid="", captchagid="-1", captcha_text="", twofactorcode=""):
        epassword, rsatimestamp = self.encrypt_password(email, password)

        login_data = {
            "username": email,
            "password": epassword,
            "emailauth": emailauth,
            "loginfriendlyname": "Streamlink",
            "captchagid": captchagid,
            "captcha_text": captcha_text,
            "emailsteamid": emailsteamid,
            "rsatimestamp": rsatimestamp,
            "remember_login": True,
            "donotcache": self.donotcache,
            "twofactorcode": twofactorcode,
        }

        resp = self.session.http.post(
            self._dologin_url,
            data=login_data,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "success": bool,
                    "requires_twofactor": bool,
                    validate.optional("message"): str,
                    validate.optional("emailauth_needed"): bool,
                    validate.optional("emaildomain"): str,
                    validate.optional("emailsteamid"): str,
                    validate.optional("login_complete"): bool,
                    validate.optional("captcha_needed"): bool,
                    validate.optional("captcha_gid"): validate.any(str, int)
                }
            )
        )

        if resp.get("login_complete"):
            return True

        if not resp["success"]:
            if resp.get("captcha_needed"):
                # special case for captcha
                captchagid = resp["captcha_gid"]
                captchaurl = self._captcha_url.format(captchagid)
                log.error(f"Captcha result required, open this URL to see the captcha: {captchaurl}")
                try:
                    captcha_text = self.input_ask("Captcha text")
                except FatalPluginError:
                    captcha_text = None
                if not captcha_text:
                    return False
            else:
                # If the user must enter the code that was emailed to them
                if resp.get("emailauth_needed"):
                    if emailauth:
                        raise SteamLoginFailed("Email auth key error")
                    try:
                        emailauth = self.input_ask("Email auth code required")
                    except FatalPluginError:
                        emailauth = None
                    if not emailauth:
                        return False

                # If the user must enter a two factor auth code
                if resp.get("requires_twofactor"):
                    try:
                        twofactorcode = self.input_ask("Two factor auth code required")
                    except FatalPluginError:
                        twofactorcode = None
                    if not twofactorcode:
                        return False

                if resp.get("message"):
                    raise SteamLoginFailed(resp["message"])

            return self.dologin(
                email,
                password,
                emailauth=emailauth,
                emailsteamid=resp.get("emailsteamid", ""),
                captcha_text=captcha_text,
                captchagid=captchagid,
                twofactorcode=twofactorcode
            )

        log.error("Something went wrong while logging in to Steam")
        return False

    def _get_broadcast_stream(self, steamid, viewertoken=0, sessionid=None):
        log.debug(f"Getting broadcast stream: sessionid={sessionid}")
        return self.session.http.get(
            self._get_broadcast_url,
            params=dict(
                broadcastid=0,
                steamid=steamid,
                viewertoken=viewertoken,
                sessionid=sessionid
            ),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "success": validate.any("ready", "unavailable", "waiting", "waiting_to_start", "waiting_for_start"),
                    "retry": int,
                    "broadcastid": validate.any(str, int),
                    validate.optional("url"): validate.url(),
                    validate.optional("viewertoken"): str
                }
            )
        )

    def _find_steamid(self, url):
        return self.session.http.get(url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//div[@id='webui_config']/@data-broadcast"),
            validate.none_or_all(
                validate.parse_json(),
                {"steamid": str},
                validate.get("steamid"),
            ),
        ))

    def _get_streams(self):
        self.session.http.headers["User-Agent"] = f"streamlink/{self.session.version}"

        email = self.get_option("email")
        if email:
            log.info(f"Attempting to login to Steam as {email}")
            try:
                success = self.dologin(email, self.get_option("password"))
            except SteamLoginFailed as err:
                log.error(err)
                return
            if success:
                log.info(f"Logged in as {email}")
                self.save_cookies(lambda c: "steamMachineAuth" in c.name)

        if self.matches[1] is None:
            steamid = self.match.group(1)
        else:
            steamid = self._find_steamid(self.url)
            if not steamid:
                return
            self.url = self._watch_broadcast_url.format(steamid=steamid)

        res = self.session.http.get(self.url)  # get the page to set some cookies
        sessionid = res.cookies.get("sessionid")

        streamdata = None
        while streamdata is None or streamdata["success"] in ("waiting", "waiting_for_start"):
            streamdata = self._get_broadcast_stream(steamid, sessionid=sessionid)

            if streamdata["success"] == "ready":
                return DASHStream.parse_manifest(self.session, streamdata["url"])

            if streamdata["success"] == "unavailable":
                log.error("This stream is currently unavailable")
                return

            r = streamdata["retry"] / 1000.0
            log.info(f"Waiting for stream, will retry again in {r:.1f} seconds...")
            time.sleep(r)


__plugin__ = SteamBroadcastPlugin



from streamlink.plugins.idf1 import IDF1
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlIDF1(PluginCanHandleUrl):
    __plugin__ = IDF1

    should_match = [
        "https://www.idf1.fr/live",
        "https://www.idf1.fr/videos/jlpp/best-of-2018-02-24-partie-2.html",
        "http://www.idf1.fr/videos/buzz-de-noel/partie-2.html",
    ]

    should_not_match = [
        "https://www.idf1.fr/",
        "https://www.idf1.fr/videos",
        "https://www.idf1.fr/programmes/emissions/idf1-chez-vous.html",
    ]



"""
$description Indian live TV channels and video on-demand service. OTT service from YuppTV.
$url yupptv.com
$type live, vod
$account Some streams require an account and subscription
"""

import logging
import re
import time

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r'https?://(?:www\.)?yupptv\.com'
))
@pluginargument(
    "boxid",
    requires=["yuppflixtoken"],
    sensitive=True,
    metavar="BOXID",
    help="The yupptv.com boxid that's used in the BoxId cookie.",
)
@pluginargument(
    "yuppflixtoken",
    sensitive=True,
    metavar="YUPPFLIXTOKEN",
    help="The yupptv.com yuppflixtoken that's used in the YuppflixToken cookie.",
)
@pluginargument(
    "purge-credentials",
    action="store_true",
    help="Purge cached YuppTV credentials to initiate a new session and reauthenticate.",
)
class YuppTV(Plugin):
    _m3u8_re = re.compile(r'''['"](http.+\.m3u8.*?)['"]''')
    _cookie_expiry = 3600 * 24 * 365

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._authed = (self.session.http.cookies.get("BoxId")
                        and self.session.http.cookies.get("YuppflixToken"))

    @staticmethod
    def _override_encoding(res, **kwargs):
        res.encoding = "utf-8"

    def _login_using_box_id_and_yuppflix_token(self, box_id, yuppflix_token):
        time_now = time.time()

        self.session.http.cookies.set(
            'BoxId',
            box_id,
            domain='www.yupptv.com',
            path='/',
            expires=time_now + self._cookie_expiry,
        )
        self.session.http.cookies.set(
            'YuppflixToken',
            yuppflix_token,
            domain='www.yupptv.com',
            path='/',
            expires=time_now + self._cookie_expiry,
        )

        self.save_cookies()
        log.info("Successfully set BoxId and YuppflixToken")

    def _get_streams(self):
        self.session.http.headers.update({"User-Agent": useragents.CHROME})
        self.session.http.headers.update({"Origin": "https://www.yupptv.com"})

        login_box_id = self.get_option("boxid")
        login_yuppflix_token = self.get_option("yuppflixtoken")

        if self.options.get("purge_credentials"):
            self.clear_cookies()
            self._authed = False
            log.info("All credentials were successfully removed")

        if self._authed:
            log.debug("Attempting to authenticate using cached cookies")
        elif login_box_id and login_yuppflix_token:
            self._login_using_box_id_and_yuppflix_token(
                login_box_id,
                login_yuppflix_token,
            )
            self._authed = True

        page = self.session.http.get(self.url)
        if self._authed and "btnsignup" in page.text:
            log.error("This device requires renewed credentials to log in")
            return

        match = self._m3u8_re.search(page.text)
        if match:
            stream_url = match.group(1)
            if "preview/" in stream_url:
                if "btnsignup" in page.text:
                    log.error("This stream requires you to login")
                else:
                    log.error("This stream requires a subscription")
                return

            return HLSStream.parse_variant_playlist(self.session,
                                                    stream_url,
                                                    hooks={"response": self._override_encoding})
        elif "btnsignup" in page.text:
            log.error("This stream requires you to login")
        elif "btnsubscribe" in page.text:
            log.error("This stream requires a subscription")


__plugin__ = YuppTV



import os
from pathlib import Path
from unittest.mock import patch

import pytest

from streamlink_cli.utils.path import replace_chars, replace_path
from tests import posix_only, windows_only


@pytest.mark.parametrize("char", list(range(32)))
def test_replace_chars_unprintable(char: int):
    assert replace_chars(f"foo{chr(char)}{chr(char)}bar") == "foo_bar", "Replaces unprintable characters"


@posix_only
@pytest.mark.parametrize("char", "/".split())
def test_replace_chars_posix(char: str):
    assert replace_chars(f"foo{char}{char}bar") == "foo_bar", "Replaces multiple unsupported characters in a row"


@windows_only
@pytest.mark.parametrize("char", "\x7f\"*/:<>?\\|".split())
def test_replace_chars_windows(char: str):
    assert replace_chars(f"foo{char}{char}bar") == "foo_bar", "Replaces multiple unsupported characters in a row"


@posix_only
def test_replace_chars_posix_all():
    assert replace_chars("".join(chr(i) for i in range(32)) + "/") == "_"


@windows_only
def test_replace_chars_windows_all():
    assert replace_chars("".join(chr(i) for i in range(32)) + "\x7f\"*/:<>?\\|") == "_"


@posix_only
def test_replace_chars_posix_override():
    all_chars = "".join(chr(i) for i in range(32)) + "\x7f\"*:/<>?\\|"
    assert replace_chars(all_chars) == "_\x7f\"*:_<>?\\|"
    assert replace_chars(all_chars, "posix") == "_\x7f\"*:_<>?\\|"
    assert replace_chars(all_chars, "unix") == "_\x7f\"*:_<>?\\|"
    assert replace_chars(all_chars, "windows") == "_"
    assert replace_chars(all_chars, "win32") == "_"


@windows_only
def test_replace_chars_windows_override():
    all_chars = "".join(chr(i) for i in range(32)) + "\x7f\"*:/<>?\\|"
    assert replace_chars(all_chars) == "_"
    assert replace_chars(all_chars, "posix") == "_\x7f\"*:_<>?\\|"
    assert replace_chars(all_chars, "unix") == "_\x7f\"*:_<>?\\|"
    assert replace_chars(all_chars, "windows") == "_"
    assert replace_chars(all_chars, "win32") == "_"


def test_replace_chars_replacement():
    assert replace_chars("\x00", None, "+") == "+"


def test_replace_path():
    def mapper(s):
        return dict(foo=".", bar="..").get(s, s)

    path = Path("foo", ".", "bar", "..", "baz")
    expected = Path("_", ".", "_", "..", "baz")
    assert replace_path(path, mapper) == expected, "Only replaces mapped parts which are in the special parts tuple"


@posix_only
def test_replace_path_expanduser_posix():
    with patch.object(os, "environ", {"HOME": "/home/foo"}):
        assert replace_path("~/bar", lambda s: s) == Path("/home/foo/bar")
        assert replace_path("foo/bar", lambda s: dict(foo="~").get(s, s)) == Path("~/bar")


@windows_only
def test_replace_path_expanduser_windows():
    with patch.object(os, "environ", {"USERPROFILE": "C:\\Users\\foo"}):
        assert replace_path("~\\bar", lambda s: s) == Path("C:\\Users\\foo\\bar")
        assert replace_path("foo\\bar", lambda s: dict(foo="~").get(s, s)) == Path("~\\bar")



import re
from io import BytesIO

from streamlink import NoStreamsError
from streamlink.options import Options
from streamlink.plugin import pluginargument, pluginmatcher
from streamlink.plugins import Plugin
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.stream.stream import Stream


class TestStream(Stream):
    __shortname__ = "test"

    def open(self):
        return BytesIO(b'x' * 8192 * 2)


@pluginmatcher(re.compile(
    r"https?://test\.se"
))
@pluginargument(
    "bool",
    action="store_true",
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
)
class TestPlugin(Plugin):
    options = Options({
        "a_option": "default"
    })

    id = "test-id-1234-5678"
    author = "Tѥst Āuƭhǿr"
    category = None
    title = "Test Title"

    def _get_streams(self):
        if "empty" in self.url:
            return

        if "UnsortableStreamNames" in self.url:
            def gen():
                for i in range(3):
                    yield "vod", HTTPStream(self.session, "http://test.se/stream")

            return gen()

        if "NoStreamsError" in self.url:
            raise NoStreamsError

        streams = {}
        streams["test"] = TestStream(self.session)
        streams["hls"] = HLSStream(self.session, "http://test.se/playlist.m3u8")
        streams["http"] = HTTPStream(self.session, "http://test.se/stream")

        streams["240p"] = HTTPStream(self.session, "http://test.se/stream")
        streams["360p"] = HTTPStream(self.session, "http://test.se/stream")
        streams["1080p"] = HTTPStream(self.session, "http://test.se/stream")

        streams["350k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["800k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["1500k"] = HTTPStream(self.session, "http://test.se/stream")
        streams["3000k"] = HTTPStream(self.session, "http://test.se/stream")

        streams["480p"] = [
            HTTPStream(self.session, "http://test.se/stream"),
            HLSStream(self.session, "http://test.se/playlist.m3u8")
        ]

        return streams


__plugin__ = TestPlugin



from streamlink.plugins.zengatv import ZengaTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlZengaTV(PluginCanHandleUrl):
    __plugin__ = ZengaTV

    should_match = [
        "http://www.zengatv.com/indiatoday.html",
        "http://www.zengatv.com/live/87021a6d-411e-11e2-b4c6-7071bccc85ac.html",
        "http://zengatv.com/indiatoday.html",
        "http://zengatv.com/live/87021a6d-411e-11e2-b4c6-7071bccc85ac.html",
    ]

    should_not_match = [
        "http://www.zengatv.com",
        "http://www.zengatv.com/"
    ]



"""
$description Live TV channels from TVR, a Romanian public, state-owned broadcaster.
$url tvrplus.ro
$type live
$region Romania
"""

import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?tvrplus\.ro(?:/live/.+|/?$)"
))
class TVRPlus(Plugin):
    def _get_streams(self):
        try:
            hls_url = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    re.compile(r"""(?P<q>["'])(?P<url>https?://\S+?\.m3u8\S*?)(?P=q)"""),
                    validate.get("url"),
                ),
            )
        except PluginError:
            return

        return HLSStream.parse_variant_playlist(self.session, hls_url, headers={"Referer": self.url})


__plugin__ = TVRPlus



"""
$description Japanese live streaming and video hosting platform for live video game broadcasts and individual live streams.
$url mildom.com
$type live, vod
"""

import logging
import re
from time import time
from uuid import uuid4

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import url_concat

log = logging.getLogger(__name__)


class MildomHLSStream(HLSStream):
    __shortname__ = "hls-mildom"
    expiry_time = 60 * 120

    def __init__(self, session_, api, server, token, quality, **args):
        self.session = session_
        self.api = api
        self.server = server
        self.token = token
        self.quality = quality
        self._url = self.build_hls_url()
        super().__init__(self.session, self._url, **args)
        self.expiry = time()

    def build_hls_url(self):
        if not self.server or not self.token:
            raise ValueError("server and token must be set")
        return url_concat(self.server, f"{self.api.channel_id}{self.quality}.m3u8?{self.token}")

    @property
    def url(self):
        if time() - self.expiry > MildomHLSStream.expiry_time:
            self.expiry = time()
            self.token = self.api.get_token()
            self._url = self.build_hls_url()
            log.debug("Updated HLS playlist URL query string")
        return self._url


class MildomAPI:
    def __init__(self, session, channel_id=None, video_id=None):
        self.session = session
        self.channel_id = channel_id
        self.video_id = video_id

    def _is_api_error(self, data):
        log.trace(f"{data!r}")
        if data["code"] != 0:
            log.debug(data.get("message", "Mildom API returned an error"))
            return True
        return False

    def get_vod_streams_data(self):
        if not self.video_id:
            return
        data = self.session.http.get(
            "https://cloudac.mildom.com/nonolive/videocontent/playback/getPlaybackDetail",
            params={
                "__platform": "web",
                "v_id": self.video_id,
            },
            schema=validate.Schema(validate.parse_json(), {
                "code": int,
                validate.optional("message"): str,
                validate.optional("body"): {
                    "playback": {
                        "video_link": [{"name": str, "url": validate.url()}],
                    },
                },
            })
        )
        if self._is_api_error(data):
            return
        if data.get("body"):
            return data["body"]["playback"]["video_link"]

    def get_token(self):
        if not self.channel_id:
            return
        data = self.session.http.post(
            "https://cloudac.mildom.com/nonolive/gappserv/live/token",
            params={
                "__platform": "web",
                "__guest_id": "pc-gp-{}".format(uuid4()),
            },
            headers={"Accept-Language": "en"},
            json={"host_id": self.channel_id, "type": "hls"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "code": int,
                    validate.optional("message"): str,
                    validate.optional("body"): {
                        "data": [
                            {"token": str, }
                        ],
                    }
                }
            )
        )
        if self._is_api_error(data):
            return
        if data.get("body"):
            return data["body"]["data"][0]["token"]

    def get_server(self):
        if not self.channel_id:
            return
        data = self.session.http.get(
            "https://cloudac.mildom.com/nonolive/gappserv/live/liveserver",
            params={
                "__platform": "web",
                "user_id": self.channel_id,
                "live_server_type": "hls",
            },
            headers={"Accept-Language": "en"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "code": int,
                    validate.optional("message"): str,
                    validate.optional("body"): {
                        "stream_server": validate.url(),
                    }
                }
            )
        )
        if self._is_api_error(data):
            return
        if data.get("body"):
            return data["body"]["stream_server"]

    def get_live_streams_data(self):
        if not self.channel_id:
            return
        data = self.session.http.get(
            "https://cloudac.mildom.com/nonolive/gappserv/live/enterstudio",
            params={
                "__platform": "web",
                "user_id": self.channel_id,
            },
            headers={"Accept-Language": "en"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "code": int,
                    validate.optional("message"): str,
                    validate.optional("body"): {
                        validate.optional("status"): int,
                        "anchor_live": int,
                        validate.optional("live_type"): int,
                        "ext": {
                            "cmode_params": [{
                                "cmode": str,
                                "name": str,
                            }],
                            validate.optional("live_mode"): int,
                        },
                    },
                },
            )
        )
        if self._is_api_error(data):
            return
        if data.get("body"):
            return data["body"]


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?mildom\.com/
    (?:
        playback/(\d+)(/(?P<video_id>(\d+)-(\w+)))
        |
        (?P<channel_id>\d+)
    )
""", re.VERBOSE))
class Mildom(Plugin):
    def _get_streams(self):
        api = MildomAPI(self.session, channel_id=self.match.group("channel_id"), video_id=self.match.group("video_id"))

        if api.video_id:
            data = api.get_vod_streams_data()
            if data:
                for stream in data:
                    yield stream["name"], HLSStream(self.session, stream["url"])
        else:
            data = api.get_live_streams_data()
            if not data:
                return

            if data["anchor_live"] != 11:
                log.debug("User doesn't appear to be live")
                return

            qualities = [
                (quality_info["name"], f"_{quality_info['cmode']}" if quality_info["cmode"] != "raw" else "")
                for quality_info in data["ext"]["cmode_params"]
            ]

            server = api.get_server()
            token = api.get_token()
            self.session.http.headers.update({"Referer": "https://www.mildom.com/"})
            for quality in qualities:
                yield quality[0], MildomHLSStream(self.session, api, server, token, quality[1])


__plugin__ = Mildom



from streamlink.plugins.cnews import CNEWS
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCNEWS(PluginCanHandleUrl):
    __plugin__ = CNEWS

    should_match = [
        "http://www.cnews.fr/le-direct",
        "http://www.cnews.fr/direct",
        "http://www.cnews.fr/emission/2018-06-12/meteo-du-12062018-784730",
        "http://www.cnews.fr/emission/2018-06-12/le-journal-des-faits-divers-du-12062018-784704"
    ]



"""
$description Global live-streaming platform for the creative community.
$url piczel.tv
$type live
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r"https?://piczel\.tv/watch/(?P<channel>\w+)"
))
class Piczel(Plugin):
    _URL_STREAMS = "https://piczel.tv/api/streams"
    _URL_HLS = "https://piczel.tv/hls/{id}/index.m3u8"

    def _get_streams(self):
        channel = self.match.group("channel")

        data = self.session.http.get(
            self._URL_STREAMS,
            params={
                "followedStreams": "false",
                "live_only": "false",
                "sfw": "false",
            },
            schema=validate.Schema(
                validate.parse_json(),
                [{
                    "slug": str,
                    "live": bool,
                    "id": int,
                    "username": str,
                    "title": str,
                }],
                validate.filter(lambda item: item["slug"] == channel),
                validate.get(0),
                validate.any(None, validate.union_get(
                    "id",
                    "username",
                    "title",
                    "live",
                )),
            ),
        )
        if not data:
            return

        self.id, self.author, self.title, is_live = data
        if not is_live:
            return

        return {"live": HLSStream(self.session, self._URL_HLS.format(id=self.id))}


__plugin__ = Piczel



import datetime
import itertools
import unittest
from operator import attrgetter
from unittest.mock import Mock

from freezegun import freeze_time
from freezegun.api import FakeDatetime  # type: ignore[attr-defined]

from streamlink.stream.dash_manifest import MPD, MPDParsers, MPDParsingError, Representation, utc
from tests.resources import xml


class TestMPDParsers(unittest.TestCase):
    def test_utc(self):
        self.assertEqual(utc.tzname(None), "UTC")
        self.assertEqual(utc.dst(None), datetime.timedelta(0))
        self.assertEqual(utc.utcoffset(None), datetime.timedelta(0))

    def test_bool_str(self):
        self.assertEqual(MPDParsers.bool_str("true"), True)
        self.assertEqual(MPDParsers.bool_str("TRUE"), True)
        self.assertEqual(MPDParsers.bool_str("True"), True)

        self.assertEqual(MPDParsers.bool_str("0"), False)
        self.assertEqual(MPDParsers.bool_str("False"), False)
        self.assertEqual(MPDParsers.bool_str("false"), False)
        self.assertEqual(MPDParsers.bool_str("FALSE"), False)

    def test_type(self):
        self.assertEqual(MPDParsers.type("dynamic"), "dynamic")
        self.assertEqual(MPDParsers.type("static"), "static")
        with self.assertRaises(MPDParsingError):
            MPDParsers.type("other")

    def test_duration(self):
        self.assertEqual(MPDParsers.duration("PT1S"), datetime.timedelta(0, 1))

    def test_datetime(self):
        self.assertEqual(MPDParsers.datetime("2018-01-01T00:00:00Z"),
                         datetime.datetime(2018, 1, 1, 0, 0, 0, tzinfo=utc))

    def test_segment_template(self):
        self.assertEqual(MPDParsers.segment_template("$Time$-$Number$-$Other$")(Time=1, Number=2, Other=3),
                         "1-2-3")
        self.assertEqual(MPDParsers.segment_template("$Number%05d$")(Number=123),
                         "00123")
        self.assertEqual(MPDParsers.segment_template("$Time%0.02f$")(Time=100.234),
                         "100.23")

    def test_frame_rate(self):
        self.assertAlmostEqual(MPDParsers.frame_rate("1/25"),
                               1 / 25.0)
        self.assertAlmostEqual(MPDParsers.frame_rate("0.2"),
                               0.2)

    def test_timedelta(self):
        self.assertEqual(MPDParsers.timedelta(1)(100),
                         datetime.timedelta(0, 100.0))
        self.assertEqual(MPDParsers.timedelta(10)(100),
                         datetime.timedelta(0, 10.0))

    def test_range(self):
        self.assertEqual(MPDParsers.range("100-"), (100, None))
        self.assertEqual(MPDParsers.range("100-199"), (100, 100))
        self.assertRaises(MPDParsingError, MPDParsers.range, "100")


class TestMPDParser(unittest.TestCase):
    maxDiff = None

    def test_segments_number_time(self):
        with xml("dash/test_1.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/tracks-v3/init-1526842800.g_m4v")

            video_segments = list(map(attrgetter("url"), (itertools.islice(segments, 5))))
            # suggested delay is 11 seconds, each segment is 5 seconds long - so there should be 3
            self.assertSequenceEqual(video_segments,
                                     ['http://test.se/tracks-v3/dvr-1526842800-698.g_m4v?t=3403000',
                                      'http://test.se/tracks-v3/dvr-1526842800-699.g_m4v?t=3408000',
                                      'http://test.se/tracks-v3/dvr-1526842800-700.g_m4v?t=3413000'])

    def test_segments_static_number(self):
        with xml("dash/test_2.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[3].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/video/250kbit/init.mp4")

            video_segments = list(map(attrgetter("url"), (itertools.islice(segments, 100000))))
            self.assertEqual(len(video_segments), 444)
            self.assertSequenceEqual(video_segments[:5],
                                     ['http://test.se/video/250kbit/segment_1.m4s',
                                      'http://test.se/video/250kbit/segment_2.m4s',
                                      'http://test.se/video/250kbit/segment_3.m4s',
                                      'http://test.se/video/250kbit/segment_4.m4s',
                                      'http://test.se/video/250kbit/segment_5.m4s'])

    def test_segments_dynamic_time(self):
        with xml("dash/test_3.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/video-2800000-0.mp4?z32=")

            video_segments = list(map(attrgetter("url"), (itertools.islice(segments, 3))))
            # default suggested delay is 3 seconds, each segment is 4 seconds long - so there should be 1 segment
            self.assertSequenceEqual(video_segments,
                                     ['http://test.se/video-time=1525450872000-2800000-0.m4s?z32='])

    def test_segments_dynamic_number(self):
        with freeze_time(FakeDatetime(2018, 5, 22, 13, 37, 0, tzinfo=utc)):
            with xml("dash/test_4.mpd") as mpd_xml:
                mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

                segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
                init_segment = next(segments)
                self.assertEqual(init_segment.url, "http://test.se/hd-5-init.mp4")

                video_segments = []
                for _ in range(3):
                    seg = next(segments)
                    video_segments.append((seg.url,
                                           seg.available_at))

                self.assertSequenceEqual(video_segments,
                                         [('http://test.se/hd-5_000311235.mp4',
                                           datetime.datetime(2018, 5, 22, 13, 37, 0, tzinfo=utc)),
                                          ('http://test.se/hd-5_000311236.mp4',
                                           datetime.datetime(2018, 5, 22, 13, 37, 5, tzinfo=utc)),
                                          ('http://test.se/hd-5_000311237.mp4',
                                           datetime.datetime(2018, 5, 22, 13, 37, 10, tzinfo=utc))
                                          ])

    def test_segments_static_no_publish_time(self):
        with xml("dash/test_5.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[1].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/dash/150633-video_eng=194000.dash")

            video_segments = [x.url for x in itertools.islice(segments, 3)]
            self.assertSequenceEqual(video_segments,
                                     ['http://test.se/dash/150633-video_eng=194000-0.dash',
                                      'http://test.se/dash/150633-video_eng=194000-2000.dash',
                                      'http://test.se/dash/150633-video_eng=194000-4000.dash',
                                      ])

    def test_segments_list(self):
        with xml("dash/test_7.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/chunk_ctvideo_ridp0va0br4332748_cinit_mpd.m4s")

            video_segments = [x.url for x in itertools.islice(segments, 3)]
            self.assertSequenceEqual(video_segments,
                                     ['http://test.se/chunk_ctvideo_ridp0va0br4332748_cn1_mpd.m4s',
                                      'http://test.se/chunk_ctvideo_ridp0va0br4332748_cn2_mpd.m4s',
                                      'http://test.se/chunk_ctvideo_ridp0va0br4332748_cn3_mpd.m4s',
                                      ])

    def test_segments_dynamic_timeline_continue(self):
        with xml("dash/test_6_p1.mpd") as mpd_xml_p1:
            with xml("dash/test_6_p2.mpd") as mpd_xml_p2:
                mpd_p1 = MPD(mpd_xml_p1, base_url="http://test.se/", url="http://test.se/manifest.mpd")

                segments_p1 = mpd_p1.periods[0].adaptationSets[0].representations[0].segments()
                init_segment = next(segments_p1)
                self.assertEqual(init_segment.url, "http://test.se/video/init.mp4")

                video_segments_p1 = [x.url for x in itertools.islice(segments_p1, 100)]
                self.assertSequenceEqual(video_segments_p1,
                                         ['http://test.se/video/1006000.mp4',
                                          'http://test.se/video/1007000.mp4',
                                          'http://test.se/video/1008000.mp4',
                                          'http://test.se/video/1009000.mp4',
                                          'http://test.se/video/1010000.mp4'])

                # Continue in the next manifest
                mpd_p2 = MPD(mpd_xml_p2,
                             base_url=mpd_p1.base_url,
                             url=mpd_p1.url,
                             timelines=mpd_p1.timelines)

                segments_p2 = mpd_p2.periods[0].adaptationSets[0].representations[0].segments(init=False)
                video_segments_p2 = [x.url for x in itertools.islice(segments_p2, 100)]
                self.assertSequenceEqual(video_segments_p2,
                                         ['http://test.se/video/1011000.mp4',
                                          'http://test.se/video/1012000.mp4',
                                          'http://test.se/video/1013000.mp4',
                                          'http://test.se/video/1014000.mp4',
                                          'http://test.se/video/1015000.mp4'])

    def test_tsegment_t_is_none_1895(self):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/1895
        """
        with xml("dash/test_8.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            segments = mpd.periods[0].adaptationSets[0].representations[0].segments()
            init_segment = next(segments)
            self.assertEqual(init_segment.url, "http://test.se/video-2799000-0.mp4?z32=CENSORED_SESSION")

            video_segments = [x.url for x in itertools.islice(segments, 3)]
            self.assertSequenceEqual(video_segments,
                                     ['http://test.se/video-time=0-2799000-0.m4s?z32=CENSORED_SESSION',
                                      'http://test.se/video-time=4000-2799000-0.m4s?z32=CENSORED_SESSION',
                                      'http://test.se/video-time=8000-2799000-0.m4s?z32=CENSORED_SESSION',
                                      ])

    def test_bitrate_rounded(self):
        def mock_rep(bandwidth):
            node = Mock(
                tag="Representation",
                attrib={
                    "id": "test",
                    "bandwidth": bandwidth,
                    "mimeType": "video/mp4"
                }
            )
            node.findall.return_value = []
            return Representation(node)

        self.assertEqual(mock_rep(1.2 * 1000.0).bandwidth_rounded, 1.2)
        self.assertEqual(mock_rep(45.6 * 1000.0).bandwidth_rounded, 46.0)
        self.assertEqual(mock_rep(134.0 * 1000.0).bandwidth_rounded, 130.0)
        self.assertEqual(mock_rep(1324.0 * 1000.0).bandwidth_rounded, 1300.0)

    def test_duplicated_resolutions(self):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/3365
        """
        with xml("dash/test_10.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")

            representations_0 = mpd.periods[0].adaptationSets[0].representations[0]
            self.assertEqual(representations_0.height, 804)
            self.assertEqual(representations_0.bandwidth, 10000.0)
            representations_1 = mpd.periods[0].adaptationSets[0].representations[1]
            self.assertEqual(representations_1.height, 804)
            self.assertEqual(representations_1.bandwidth, 8000.0)

    def test_segments_static_periods_duration(self):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/2873
        """
        with xml("dash/test_11_static.mpd") as mpd_xml:
            mpd = MPD(mpd_xml, base_url="http://test.se/", url="http://test.se/manifest.mpd")
            duration = mpd.periods[0].duration.total_seconds()
            self.assertEqual(duration, 204.32)



from streamlink.plugins.crunchyroll import Crunchyroll
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCrunchyroll(PluginCanHandleUrl):
    __plugin__ = Crunchyroll

    should_match_groups = [
        ("http://www.crunchyroll.com/idol-incidents/episode-1-a-title-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.com/ru/idol-incidents/episode-1-a-title-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.com/idol-incidents/media-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.com/fr/idol-incidents/media-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.com/media-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.com/de/media-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.fr/media-728233", {"media_id": "728233"}),
        ("http://www.crunchyroll.fr/es/media-728233", {"media_id": "728233"}),
        ("https://beta.crunchyroll.com/watch/GRNQ5DDZR/Game-Over", {"beta_id": "GRNQ5DDZR"}),
        ("https://beta.crunchyroll.com/watch/ValidID123/any/thing?x&y", {"beta_id": "ValidID123"}),
    ]

    should_not_match = [
        "http://www.crunchyroll.com/gintama",
        "http://www.crunchyroll.es/gintama",
        "http://beta.crunchyroll.com/",
        "http://beta.crunchyroll.com/something",
        "http://beta.crunchyroll.com/watch/",
        "http://beta.crunchyroll.com/watch/not-a-valid-id",
        "http://beta.crunchyroll.com/watch/not-a-valid-id/a-title",
    ]



from streamlink.plugins.afreeca import AfreecaTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlAfreecaTV(PluginCanHandleUrl):
    __plugin__ = AfreecaTV

    should_match = [
        "http://play.afreecatv.com/exampleuser",
        "http://play.afreecatv.com/exampleuser/123123123",
        "https://play.afreecatv.com/exampleuser",
    ]

    should_not_match = [
        "http://afreeca.com/exampleuser",
        "http://afreeca.com/exampleuser/123123123",
        "http://afreecatv.com/exampleuser",
        "http://afreecatv.com/exampleuser/123123123",
        "http://www.afreecatv.com.tw",
        "http://www.afreecatv.jp",
    ]



from streamlink.plugins.galatasaraytv import GalatasarayTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlGalatasarayTV(PluginCanHandleUrl):
    __plugin__ = GalatasarayTV

    should_match = [
        'http://galatasaray.com/',
        'https://galatasaray.com',
        'https://galatasaray.com/',
        'https://www.galatasaray.com/',
    ]



from streamlink.plugins.lnk import LNK
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlLNK(PluginCanHandleUrl):
    __plugin__ = LNK

    should_match = [
        "https://lnk.lt/tiesiogiai",
        "https://lnk.lt/tiesiogiai#lnk",
        "https://lnk.lt/tiesiogiai#btv",
        "https://lnk.lt/tiesiogiai#2tv",
        "https://lnk.lt/tiesiogiai#infotv",
        "https://lnk.lt/tiesiogiai#tv1",
    ]

    should_not_match = [
        "https://lnk.lt/",
        "https://lnk.lt/vaikams",
        "https://lnk.lt/zinios/Visi/157471",
    ]



"""
$description Anime video on-demand service.
$url funimation.com
$url funimationnow.uk
$type vod
$notes :ref:`Requires session cookies <cli/plugins/funimationnow:Authentication>`
"""

import logging
import random
import re

from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.utils.l10n import Localization

log = logging.getLogger(__name__)


class Experience:
    CSRF_NAME = "csrfmiddlewaretoken"
    login_url = "https://www.funimation.com/log-in/"
    api_base = "https://www.funimation.com/api"
    login_api_url = "https://prod-api-funimationnow.dadcdigital.com/api/auth/login/"
    show_api_url = api_base + "/experience/{experience_id}/"
    sources_api_url = api_base + "/showexperience/{experience_id}/"
    languages = ["english", "japanese"]
    alphas = ["uncut", "simulcast"]

    login_schema = validate.Schema(validate.any(
        {"success": False,
         "error": str},
        {"token": str,
         "user": {"id": int}}
    ))

    def __init__(self, session, experience_id):
        """
        :param session: streamlink session
        :param experience_id: starting experience_id, may be changed later
        """
        self.session = session
        self.experience_id = experience_id
        self._language = None
        self.cache = {}
        self.token = None

    def request(self, method, url, *args, **kwargs):
        headers = kwargs.pop("headers", {})
        if self.token:
            headers.update({"Authorization": "Token {0}".format(self.token)})
            self.session.http.cookies.update({"src_token": self.token})

        log.debug("Making {0}request to {1}".format("authorized " if self.token else "", url))

        res = self.session.http.request(method, url, *args, headers=headers, **kwargs)
        if "_Incapsula_Resource" in res.text:
            log.error(
                "This page is protected by Incapsula, please see "
                "https://github.com/streamlink/streamlink/issues/2088"
                " for a workaround."
            )
            return
        return res

    def get(self, *args, **kwargs):
        return self.request("GET", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request("POST", *args, **kwargs)

    @property
    def pinst_id(self):
        return ''.join([
            random.choice("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ") for _ in range(8)
        ])

    def _update(self):
        api_url = self.show_api_url.format(experience_id=self.experience_id)
        log.debug("Requesting experience data: {0}".format(api_url))
        res = self.get(api_url)
        if res:
            data = self.session.http.json(res)
            self.cache[self.experience_id] = data

    @property
    def show_info(self):
        if self.experience_id not in self.cache:
            self._update()
        return self.cache.get(self.experience_id)

    @property
    def episode_info(self):
        """
        Search for the episode with the requested experience Id
        :return:
        """
        if self.show_info:
            for season in self.show_info["seasons"]:
                for episode in season["episodes"]:
                    for lang in episode["languages"].values():
                        for alpha in lang["alpha"].values():
                            if alpha["experienceId"] == self.experience_id:
                                return episode

    @property
    def language(self):
        for language, lang_data in self.episode_info["languages"].items():
            for alpha in lang_data["alpha"].values():
                if alpha["experienceId"] == self.experience_id:
                    return language

    @property
    def language_code(self):
        return {"english": "eng", "japanese": "jpn"}[self.language]

    def set_language(self, language):
        if language in self.episode_info["languages"]:
            for alpha in self.episode_info["languages"][language]["alpha"].values():
                self.experience_id = alpha["experienceId"]

    def _get_alpha(self):
        for lang_data in self.episode_info["languages"].values():
            for alpha in lang_data["alpha"].values():
                if alpha["experienceId"] == self.experience_id:
                    return alpha

    def subtitles(self):
        alpha = self._get_alpha()
        for src in alpha["sources"]:
            return src["textTracks"]

    def sources(self):
        """
        Get the sources for a given experience_id, which is tied to a specific language
        :param experience_id: int; video content id
        :return: sources dict
        """
        api_url = self.sources_api_url.format(experience_id=self.experience_id)
        res = self.get(api_url, params={"pinst_id": self.pinst_id})
        return self.session.http.json(res)

    def login_csrf(self):
        return self.session.http.get(self.login_url, schema=validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//input[@name=$name][1]/@value", name=self.CSRF_NAME)
        ))

    def login(self, email, password):
        log.debug("Attempting to login as {0}".format(email))
        r = self.post(self.login_api_url,
                      data={'username': email, 'password': password, self.CSRF_NAME: self.login_csrf()},
                      raise_for_status=False,
                      headers={"Referer": "https://www.funimation.com/log-in/"})
        d = self.session.http.json(r, schema=self.login_schema)
        self.token = d.get("token", None)
        return self.token is not None


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?funimation(\.com|now\.uk)"
))
@pluginargument(
    "email",
    argument_name="funimation-email",
    requires=["password"],
    help="Email address for your Funimation account.",
)
@pluginargument(
    "password",
    argument_name="funimation-password",
    sensitive=True,
    help="Password for your Funimation account.",
)
@pluginargument(
    "language",
    argument_name="funimation-language",
    choices=["en", "ja", "english", "japanese"],
    default="english",
    help="""
        The audio language to use for the stream; japanese or english.

        Default is "english".
    """,
)
@pluginargument(
    "mux-subtitles",
    is_global=True,
)
class FunimationNow(Plugin):
    experience_id_re = re.compile(r"/player/(\d+)")
    mp4_quality = "480p"

    def _get_streams(self):
        self.session.http.headers = {"User-Agent": useragents.CHROME}
        res = self.session.http.get(self.url)

        # remap en to english, and ja to japanese
        rlanguage = {"en": "english", "ja": "japanese"}.get(self.get_option("language").lower(),
                                                            self.get_option("language").lower())
        if "_Incapsula_Resource" in res.text:
            log.error(
                "This page is protected by Incapsula, please see "
                "https://github.com/streamlink/streamlink/issues/2088"
                " for a workaround."
            )
            return

        if "Out of Territory" in res.text:
            log.error("The content requested is not available in your territory.")
            return

        id_m = self.experience_id_re.search(res.text)
        experience_id = id_m and int(id_m.group(1))
        if experience_id:
            log.debug(f"Found experience ID: {experience_id}")
            exp = Experience(self.session, experience_id)
            if self.get_option("email") and self.get_option("password"):
                if exp.login(self.get_option("email"), self.get_option("password")):
                    log.info(f"Logged in to Funimation as {self.get_option('email')}")
                else:
                    log.warning("Failed to login")

            if exp.episode_info:
                log.debug(f"Found episode: {exp.episode_info['episodeTitle']}")
                log.debug(f"  has languages: {', '.join(exp.episode_info['languages'].keys())}")
                log.debug(f"  requested language: {rlanguage}")
                log.debug(f"  current language:   {exp.language}")
                if rlanguage != exp.language:
                    log.debug(f"switching language to: {rlanguage}")
                    exp.set_language(rlanguage)
                    if exp.language != rlanguage:
                        log.warning(f"Requested language {rlanguage} is not available, continuing with {exp.language}")
                    else:
                        log.debug(f"New experience ID: {exp.experience_id}")

                subtitles = None
                stream_metadata = {}
                disposition = {}
                for subtitle in exp.subtitles():
                    log.debug(f"Subtitles: {subtitle['src']}")
                    if subtitle["src"].endswith(".vtt") or subtitle["src"].endswith(".srt"):
                        sub_lang = Localization.get_language(subtitle["language"]).alpha3
                        # pick the first suitable subtitle stream
                        subtitles = subtitles or HTTPStream(self.session, subtitle["src"])
                        stream_metadata["s:s:0"] = ["language={0}".format(sub_lang)]
                    stream_metadata["s:a:0"] = ["language={0}".format(exp.language_code)]

                sources = exp.sources()
                if 'errors' in sources:
                    for error in sources['errors']:
                        log.error("{0} : {1}".format(error['title'], error['detail']))
                    return

                for item in sources["items"]:
                    url = item["src"]
                    if ".m3u8" in url:
                        for q, s in HLSStream.parse_variant_playlist(self.session, url).items():
                            if self.get_option("mux_subtitles") and subtitles:
                                yield q, MuxedStream(self.session, s, subtitles, metadata=stream_metadata,
                                                     disposition=disposition)
                            else:
                                yield q, s
                    elif ".mp4" in url:
                        # TODO: fix quality
                        s = HTTPStream(self.session, url)
                        if self.get_option("mux_subtitles") and subtitles:
                            yield self.mp4_quality, MuxedStream(self.session, s, subtitles, metadata=stream_metadata,
                                                                disposition=disposition)
                        else:
                            yield self.mp4_quality, s

        else:
            log.error("Could not find experience ID?!")


__plugin__ = FunimationNow



from streamlink.plugins.openrectv import OPENRECtv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlOPENRECtv(PluginCanHandleUrl):
    __plugin__ = OPENRECtv

    should_match = [
        'https://www.openrec.tv/live/DXRLAPSGTpx',
        'https://www.openrec.tv/movie/JsDw3rAV2Rj',
    ]

    should_not_match = [
        'https://www.openrec.tv/',
    ]



"""
$description Indonesian & international live TV channels and video on-demand service. OTT service from Vidio.
$url vidio.com
$type live, vod
"""
import logging
import re
from urllib.parse import urlsplit, urlunsplit

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?vidio\.com/"
))
class Vidio(Plugin):
    tokens_url = "https://www.vidio.com/live/{id}/tokens"

    def _get_stream_token(self, stream_id, stream_type):
        log.debug("Getting stream token")
        return self.session.http.post(
            self.tokens_url.format(id=stream_id),
            params={"type": stream_type},
            headers={"Referer": self.url},
            schema=validate.Schema(
                validate.parse_json(),
                {"token": str},
                validate.get("token"),
            ),
        )

    def _get_streams(self):
        stream_id, has_token, hls_url, dash_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//*[@data-video-id]"),
                validate.union((
                    validate.get("data-video-id"),
                    validate.all(
                        validate.get("data-video-has-token"),
                        validate.transform(lambda val: val and val != "false"),
                    ),
                    validate.get("data-vjs-clip-hls-url"),
                    validate.get("data-vjs-clip-dash-url"),
                )),
            ),
        )

        if dash_url and has_token:
            token = self._get_stream_token(stream_id, "dash")
            parsed = urlsplit(dash_url)
            dash_url = urlunsplit(parsed._replace(path=f"{token}{parsed.path}"))
            return DASHStream.parse_manifest(
                self.session,
                dash_url,
                headers={"Referer": "https://www.vidio.com/"},
            )

        if not hls_url:
            return

        if has_token:
            token = self._get_stream_token(stream_id, "hls")
            hls_url = f"{hls_url}?{token}"

        return HLSStream.parse_variant_playlist(
            self.session,
            hls_url,
            headers={"Referer": "https://www.vidio.com/"},
        )


__plugin__ = Vidio



from streamlink.plugins.vtvgo import VTVgo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlVTVgo(PluginCanHandleUrl):
    __plugin__ = VTVgo

    should_match = [
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv1-1.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv2-2.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv3-3.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv4-4.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv5-5.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv6-6.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv7-27.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv8-36.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv9-39.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-vtv5-t%C3%A2y-nam-b%E1%BB%99-7.html',
        'https://vtvgo.vn/xem-truc-tuyen-kenh-k%C3%AAnh-vtv5-t%C3%A2y-nguy%C3%AAn-163.html',
    ]

    should_not_match = [
        # POST request will error with www.
        'https://www.vtvgo.vn/xem-truc-tuyen-kenh-vtv1-1.html',
    ]



from typing import TYPE_CHECKING

# noinspection PyPep8Naming,PyShadowingBuiltins
from streamlink.plugin.api.validate._schemas import (  # noqa: I101, F401
    SchemaContainer,
    AllSchema as all,
    AnySchema as any,
    NoneOrAllSchema as none_or_all,
    ListSchema as list,
    RegexSchema as regex,
    TransformSchema as transform,
    OptionalSchema as optional,
    GetItemSchema as get,
    AttrSchema as attr,
    UnionSchema as union,
    UnionGetSchema as union_get,
    XmlElementSchema as xml_element,
)
from streamlink.plugin.api.validate._validate import (  # noqa: F401
    Schema,
    validate,
)
# noinspection PyShadowingBuiltins
from streamlink.plugin.api.validate._validators import (  # noqa: I101, F401
    validator_length as length,
    validator_startswith as startswith,
    validator_endswith as endswith,
    validator_contains as contains,
    validator_url as url,
    validator_getattr as getattr,
    validator_hasattr as hasattr,
    validator_filter as filter,
    validator_map as map,
    validator_xml_find as xml_find,
    validator_xml_findall as xml_findall,
    validator_xml_findtext as xml_findtext,
    validator_xml_xpath as xml_xpath,
    validator_xml_xpath_string as xml_xpath_string,
    validator_parse_json as parse_json,
    validator_parse_html as parse_html,
    validator_parse_xml as parse_xml,
    validator_parse_qsd as parse_qsd,
)


if TYPE_CHECKING:  # pragma: no cover
    from typing import Type

    text: Type[str]


def _deprecations():
    import sys

    deprecations = {
        "text": (str, f"`{__name__}.text` is deprecated. Use `str` instead."),
    }

    def __getattr__(_attr: str):
        if _attr in deprecations:
            import warnings
            from streamlink.exceptions import StreamlinkDeprecationWarning

            val, msg = deprecations[_attr]
            warnings.warn(msg, StreamlinkDeprecationWarning)

            return val

        raise AttributeError

    __all__ = [k for k in globals().keys() if not k.startswith("_")]
    __all__.extend(deprecations.keys())

    setattr(sys.modules[__name__], "__getattr__", __getattr__)
    setattr(sys.modules[__name__], "__all__", __all__)


_deprecations()
del _deprecations



"""
$description Stream videos stored in Google Drive.
$url docs.google.com
$url drive.google.com
$type vod
"""

import logging
import re
from urllib.parse import parse_qsl

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:drive|docs)\.google\.com/file/d/([^/]+)/?"
))
class GoogleDocs(Plugin):
    api_url = "https://docs.google.com/get_video_info"

    def _get_streams(self):
        docid = self.match.group(1)
        log.debug("Google Docs ID: {0}".format(docid))
        res = self.session.http.get(self.api_url, params=dict(docid=docid))
        data = dict(parse_qsl(res.text))

        if data["status"] == "ok":
            fmts = dict([s.split('/')[:2] for s in data["fmt_list"].split(",")])
            streams = [s.split('|') for s in data["fmt_stream_map"].split(",")]
            for qcode, url in streams:
                _, h = fmts[qcode].split("x")
                yield "{0}p".format(h), HTTPStream(self.session, url)
        else:
            log.error("{0} (ID: {1})".format(data["reason"], docid))


__plugin__ = GoogleDocs



"""
Convert an argparse parser to option directives.

Inspired by sphinxcontrib.autoprogram but with a few differences:

- Contains some simple pre-processing on the help messages to make
  the Sphinx version a bit prettier.
"""

import argparse
import re
from textwrap import dedent

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged
from docutils.statemachine import ViewList
from sphinx.util.nodes import nested_parse_with_titles


_block_re = re.compile(r":\n{2}\s{2}")
_default_re = re.compile(r"Default is (.+)\.\n")
_note_re = re.compile(r"Note: (.*)(?:\n\n|\n*$)", re.DOTALL)
_option_line_re = re.compile(r"^(?!\s{2,}%\(prog\)s|\s{2,}--\w[\w-]*\w\b|Example: )(.+)$", re.MULTILINE)
_option_re = re.compile(r"(?:^|(?<=\s))(--\w[\w-]*\w)\b")
_prog_re = re.compile(r"%\(prog\)s")
_percent_re = re.compile(r"%%")
_cli_metadata_variables_section_cross_link_re = re.compile(r"the \"Metadata variables\" section")
_inline_code_block_re = re.compile(r"(?<!`)`([^`]+?)`")
_example_inline_code_block_re = re.compile(r"(?<=^Example: )(.+)$", re.MULTILINE)


def get_parser(module_name, attr):
    module = __import__(module_name, globals(), locals(), [attr])
    parser = getattr(module, attr)
    return parser if not callable(parser) else parser()


def indent(value, length=4):
    space = " " * length
    return "\n".join(space + line for line in value.splitlines())


class ArgparseDirective(Directive):
    has_content = True
    option_spec = {
        "module": unchanged,
        "attr": unchanged,
    }

    _headlines = ["^", "~"]

    def process_help(self, help):
        # Dedent the help to make sure we are always dealing with
        # non-indented text.
        help = dedent(help)

        help = _inline_code_block_re.sub(
            lambda m: (
                ":code:`{0}`".format(m.group(1).replace('\\', '\\\\'))
            ),
            help
        )

        help = _example_inline_code_block_re.sub(r":code:`\1`", help)

        # Replace option references with links.
        # Do this before indenting blocks and notes.
        help = _option_line_re.sub(
            lambda m: (
                _option_re.sub(
                    lambda m2: (
                        ":option:`{0}`".format(m2.group(1))
                        if m2.group(1) in self._available_options
                        else m2.group(0)
                    ),
                    m.group(1)
                )
            ),
            help
        )

        # Create simple blocks.
        help = _block_re.sub("::\n\n  ", help)

        # Boldify the default value.
        help = _default_re.sub(r"Default is: **\1**.\n", help)

        # Create note directives from "Note: " paragraphs.
        help = _note_re.sub(
            lambda m: ".. note::\n\n" + indent(m.group(1)) + "\n\n",
            help
        )

        # workaround to replace %(prog)s with streamlink
        help = _prog_re.sub("streamlink", help)

        # fix escaped chars for percent-formatted argparse help strings
        help = _percent_re.sub("%", help)

        # create cross-link for the "Metadata variables" section
        help = _cli_metadata_variables_section_cross_link_re.sub(
            "the \":ref:`Metadata variables <cli/metadata:Variables>`\" section",
            help
        )

        return indent(help)

    def generate_group_rst(self, group):
        for action in group._group_actions:
            # don't document suppressed parameters
            if action.help == argparse.SUPPRESS:
                continue

            metavar = action.metavar
            if isinstance(metavar, tuple):
                metavar = " ".join(metavar)

            options = []
            # parameter(s) with metavar
            if action.option_strings and metavar:
                for arg in action.option_strings:
                    # optional parameter value
                    if action.nargs == "?":
                        metavar = f"[{metavar}]"
                    options.append(f"{arg} {metavar}")
            # positional parameter
            elif metavar:
                options.append(metavar)
            # parameter(s) without metavar
            else:
                options += action.option_strings

            directive = ".. option:: "
            options = f"\n{' ' * len(directive)}".join(options)
            yield f"{directive}{options}"
            yield ""
            for line in self.process_help(action.help).split("\n"):
                yield line
            yield ""
            if hasattr(action, "plugins") and len(action.plugins) > 0:
                yield f"    **Supported plugins:** {', '.join(action.plugins)}"
                yield ""

    def generate_parser_rst(self, parser, parent=None, depth=0):
        if depth >= len(self._headlines):
            return
        for group in parser.NESTED_ARGUMENT_GROUPS[parent]:
            is_parent = group in parser.NESTED_ARGUMENT_GROUPS
            # Exclude empty groups
            if not group._group_actions and not is_parent:
                continue
            title = group.title
            yield ""
            yield title
            yield self._headlines[depth] * len(title)
            yield from self.generate_group_rst(group)
            if is_parent:
                yield ""
                yield from self.generate_parser_rst(parser, group, depth + 1)

    def run(self):
        module = self.options.get("module")
        attr = self.options.get("attr")
        parser = get_parser(module, attr)

        self._available_options = []
        for action in parser._actions:
            # positional parameters have an empty option_strings list
            self._available_options += action.option_strings or [action.dest]

        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        for line in self.generate_parser_rst(parser):
            result.append(line, "argparse")

        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    app.add_directive("argparse", ArgparseDirective)



"""
$description Live TV channels from BTRC, a Belarusian public, state-owned broadcaster.
$url tvr.by
$type live
$region Belarus
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?tvr\.by/televidenie/belarus"
))
class TVRBy(Plugin):
    file_re = re.compile(r"""(?P<url>https://stream\.hoster\.by[^"',]+\.m3u8[^"',]*)""")
    player_re = re.compile(r"""["'](?P<url>[^"']+tvr\.by/plugines/online-tv-main\.php[^"']+)["']""")

    stream_schema = validate.Schema(
        validate.all(
            validate.transform(file_re.finditer),
            validate.transform(list),
            [validate.get("url")],
            # remove duplicates
            validate.transform(set),
            validate.transform(list),
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # ensure the URL ends with a /
        if not self.url.endswith("/"):
            self.url += "/"

    def _get_streams(self):
        res = self.session.http.get(self.url)
        m = self.player_re.search(res.text)
        if not m:
            return

        player_url = m.group("url")
        res = self.session.http.get(player_url)
        stream_urls = self.stream_schema.validate(res.text)
        log.debug("Found {0} stream URL{1}".format(len(stream_urls), "" if len(stream_urls) == 1 else "s"))

        for stream_url in stream_urls:
            yield from HLSStream.parse_variant_playlist(self.session, stream_url).items()


__plugin__ = TVRBy
